# -*- coding: utf-8 -*-

import collections
import datetime
import logging
import random
import uuid

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from blinker import signal


def mysql_pub(mysql_dsn, tables=None, blocking=True, server_id=None, **kwargs):
    """MySQL row-based binlog events publisher.

    The additional kwargs will be passed to `BinLogStreamReader`.
    """
    # only import when used to avoid dependency requirements
    import pymysqlreplication
    from pymysqlreplication.row_event import (
        DeleteRowsEvent,
        UpdateRowsEvent,
        WriteRowsEvent,
    )

    logger = logging.getLogger("meepo.pub.mysql_pub")

    # parse mysql settings
    parsed = urlparse(mysql_dsn)
    mysql_settings = {
        "host": parsed.hostname,
        "port": parsed.port or 3306,
        "user": parsed.username,
        "passwd": parsed.password
    }
    server_id = server_id or random.randint(1000000000, 4294967295)

    # connect to binlog stream
    stream = pymysqlreplication.BinLogStreamReader(
        connection_settings=mysql_settings,
        blocking=blocking,
        server_id=server_id,
        only_events=[DeleteRowsEvent, UpdateRowsEvent, WriteRowsEvent],
        **kwargs
    )

    def _pk(values):
        if isinstance(event.primary_key, str):
            return values[event.primary_key]
        return tuple(values[k] for k in event.primary_key)

    for event in stream:
        if tables and event.table not in tables:
            continue

        try:
            rows = event.rows
        except (UnicodeDecodeError, ValueError) as e:
            logger.exception(e)
            continue

        timestamp = datetime.datetime.fromtimestamp(event.timestamp)

        if isinstance(event, WriteRowsEvent):
            sg_name = "{}_write".format(event.table)
            sg = signal(sg_name)

            for row in rows:
                pk = _pk(row["values"])
                sg.send(pk)

                logger.debug("{} -> {}, {}".format(sg_name, pk, timestamp))

        elif isinstance(event, UpdateRowsEvent):
            sg_name = "{}_update".format(event.table)
            sg = signal(sg_name)

            for row in rows:
                pk = _pk(row["after_values"])
                sg.send(pk)

                logger.debug("{} -> {}, {}".format(sg_name, pk, timestamp))

        elif isinstance(event, DeleteRowsEvent):
            sg_name = "{}_delete".format(event.table)
            sg = signal(sg_name)

            for row in rows:
                pk = _pk(row["values"])
                sg.send(pk)

                logger.debug("{} -> {}, {}".format(sg_name, pk, timestamp))

        signal("mysql_binlog_pos").send("{}:{}".format(stream.log_file,
                                                       stream.log_pos))


def sqlalchemy_pub(dbsession, strict_tables=None):
    """SQLAlchemy events publisher.

    This publisher don't publish events itself, it will hook on sqlalchemy's
    event system, and publish the changes automatically.

    When `strict_tables` provided, two more signals may be triggered for table
    included. They can be used with prepare-commit pattern to help ensure 100%
    reliablity on event sourcing (and broadcasting), note this will
    need another monitor daemon to monitor the status. For more info about
    this patter, refer to documentation.
    """
    from sqlalchemy import event

    logger = logging.getLogger("meepo.pub.sqlalchemy_pub")

    def _pk(obj):
        """Get pk values from object
        """
        pk_values = tuple(getattr(obj, c.name)
                          for c in obj.__mapper__.primary_key)
        if len(pk_values) == 1:
            return pk_values[0]
        return pk_values

    def _pub(obj, action):
        """Publish object pk values with action.
        """
        sg_name = "{}_{}".format(obj.__table__, action)
        sg = signal(sg_name)

        pk = _pk(obj)
        if pk:
            logger.debug("{} -> {}".format(sg_name, pk))
            sg.send(pk)

    def _init_session(session):
        for action in ("write", "update", "delete"):
            attr = "pending_{}".format(action)
            if not hasattr(session, attr):
                setattr(session, attr, set())

    def after_flush_hook(session, flush_ctx):
        """Record session changes on flush
        """
        _init_session(session)
        session.pending_write |= set(session.new)
        session.pending_update |= set(session.dirty)
        session.pending_delete |= set(session.deleted)
    event.listen(dbsession, "after_flush", after_flush_hook)

    def _pub_session(session):
        _init_session(session)

        for obj in session.pending_write:
            _pub(obj, action="write")
        for obj in session.pending_update:
            _pub(obj, action="update")
        for obj in session.pending_delete:
            _pub(obj, action="delete")

        session.pending_write.clear()
        session.pending_update.clear()
        session.pending_delete.clear()

    if not strict_tables:
        def after_commit_hook(session):
            """Publish signals
            """
            _pub_session(session)
        event.listen(dbsession, "after_commit", after_commit_hook)

    else:
        logger.debug("strict_tables: {}".format(strict_tables))

        def _strict_filter(objs):
            return (obj for obj in objs
                    if obj.__table__.fullname in strict_tables)

        def session_prepare(session):
            """Record session prepare state in before_commit
            """
            assert not hasattr(session, 'meepo_unique_id')
            _init_session(session)
            session.meepo_unique_id = uuid.uuid4().hex
            for action in ("write", "update", "delete"):
                objs = [o for o in getattr(session, "pending_%s" % action)
                        if o.__table__.fullname in strict_tables]
                if not objs:
                    continue

                prepare_event = collections.defaultdict(set)
                for obj in objs:
                    prepare_event[obj.__table__.fullname].add(_pk(obj))
                logger.debug("session_prepare {}: {} -> {}".format(
                    action, session.meepo_unique_id, prepare_event))
                signal("session_prepare").send(
                    prepare_event, sid=session.meepo_unique_id, action=action)
        event.listen(dbsession, "before_commit", session_prepare)

        def session_commit(session):
            """Commit session in after_commit
            """
            assert hasattr(session, 'meepo_unique_id')

            # normal session pub
            _pub_session(session)

            logger.debug("session_commit: {}".format(session.meepo_unique_id))
            signal("session_commit").send(session.meepo_unique_id)
            del session.meepo_unique_id
        event.listen(dbsession, "after_commit", session_commit)

        def session_rollback(session):
            """Unprepare session in after_rollback.
            """
            if not hasattr(session, 'meepo_unique_id'):
                return

            logger.debug("session_rollback: {}".format(
                session.meepo_unique_id))
            signal("session_rollback").send(session.meepo_unique_id)
            del session.meepo_unique_id
        event.listen(dbsession, "after_rollback", session_rollback)



# -*- coding: utf-8 -*-

"""
    meepo.apps.mreplicate
    ~~~~~~~~~~~~~~~~~~~~~

    Database replication app.

    This replication is a limited row-based write/update/delete replication.
    You have to make sure the schema was created and matched between master
    and slave before the replication starts.

    Currently, only replicate from mysql master with row-based binlog is
    supported.
"""

import logging
logger = logging.getLogger(__name__)

import click

from meepo.pub import mysql_pub
from meepo.sub import replicate_sub


@click.command()
@click.option('-m', '--master_dsn')
@click.option('-s', '--slave_dsn')
@click.argument('tables', nargs=-1)
def main(master_dsn, slave_dsn, tables):
    # currently only supports mysql master
    assert master_dsn.startswith("mysql")

    if tables:
        logger.info("replicating tables: %s" % ", ".join(tables))
    else:
        tables = None

    replicate_sub(master_dsn, slave_dsn, tables)
    mysql_pub(master_dsn)



# -*- coding: utf-8 -*-

"""
    meepo.apps.mzdevice
    ~~~~~~~~~~~~~~~~~~~

    ZeroMQ Forwarder Device.

    This is a forwarder that connect multiple pubs and subs.

    Refer to http://learning-0mq-with-pyzmq.readthedocs.org/en/latest/pyzmq/devices/forwarder.html  # NOQA
"""

import logging

import click
import zmq

from meepo.logutils import setup_logger


@click.command()
@click.option('-f', '--frontend_bind')
@click.option('-b', '--backend_bind')
def main(frontend_bind, backend_bind):
    setup_logger("INFO")

    logger = logging.getLogger(__name__)

    assert frontend_bind and backend_bind

    try:
        ctx = zmq.Context()
        frontend = ctx.socket(zmq.SUB)
        frontend.bind(frontend_bind)
        frontend.setsockopt(zmq.SUBSCRIBE, b"")

        backend = ctx.socket(zmq.PUB)
        backend.bind(backend_bind)

        zmq.device(zmq.FORWARDER, frontend, backend)

    except Exception as e:
        logger.exception(e)

    finally:
        frontend.close()
        backend.close()



# -*- coding: utf-8 -*-

"""
    meepo.replicator
    ~~~~~~~~~~~~~~~~

    Meepo Replicators based on events.

    Replicate from database to targets.
"""

from multiprocessing import Process, Queue

import zmq
from zmq.utils.strtypes import asbytes


def replicate(topic, pk_queue, callback):
    while True:
        pk = pk_queue.get()
        callback(topic, pk)


class Worker(Process):
    def __init__(self, queue, cb):
        super(Worker, self).__init__()
        self.queue = queue
        self.cb = cb

    def run(self):
        for pk in iter(self.queue.get, None):
            self.cb(pk)


class ZmqReplicator(object):
    """Replicator base class.

    Trigger retrive, process, store steps to do replication.
    """

    def __init__(self, listen):
        self.listen = listen
        self.workers = {}
        self.worker_queues = {}

        # init zmq socket
        self._ctx = zmq.Context()
        self.socket = self._ctx.socket(zmq.SUB)

    def event(self, topic):
        """Topic callback registry.

        callback func should receive two args: topic and pk, and then process
        the replication job.
        """
        def wrapper(func):
            self.worker_queues[topic] = Queue()
            self.workers[topic] = Worker(self.worker_queues[topic], func)
            self.socket.setsockopt(zmq.SUBSCRIBE, asbytes(topic))
            return func
        return wrapper

    def run(self):
        """Run replicator.

        Main process receive messages and distribute them to worker queues.
        """
        for w in self.workers.values():
            w.start()

        self.socket.connect(self.listen)
        while True:
            msg = self.socket.recv_string()
            topic, pk = msg.split()
            self.worker_queues[topic].put(pk)
